//#define _USE_32BIT_TIME_T
//#define BUILDING_DLL 1


#include <malloc.h>
#include <stdio.h>
#include <postgres.h>
#include <fmgr.h>

#include "numeric2.h"

PG_MODULE_MAGIC;

////////////////////////////////////////////////////////////////////////////
////////////////////// Satic values for speed up ///////////////////////////
char* radix_characters = "0123456789ABCDEF";

//inline size_t calculeteNumericSize(int numberOfComponents);

Size calculeteNumericSize(int numberOfComponents) {
	return sizeof(numeric2cmp)*numberOfComponents + sizeof(_numeric2) + 12;
}

numeric2 numeric2_allocate(short numberOfComponents, short precision, short scale) {
	int i;
	int* values;

	numeric2 result = (numeric2) palloc0(calculeteNumericSize(numberOfComponents));
	SET_VARSIZE(result, calculeteNumericSize(numberOfComponents));

	result->size = calculeteNumericSize(numberOfComponents);
	result->precision = precision;
	result->scale = scale;
	strcpy(((void*) result) + sizeof(_numeric2) + numberOfComponents*sizeof(int), "__numeric2__");
	values = getNumericValues(result);
	fprintf(stderr, "Stored string is %s\n", ((void*) result) + sizeof(_numeric2) + numberOfComponents*sizeof(int));
	//Calloc array;
	for (i=0; i < getNumericValuesLength(result); i++)
		values[i] = 0;
	return result;
}

numeric2 numeric2_reallocate(numeric2 n, short newNumberOfComponents) {
	fprintf(stderr, "Realocating numeric\n");
	numeric2 out;
	short oldSize;

	if (n->size == newNumberOfComponents)
		return n;
	if (n->size < newNumberOfComponents) {
		out = (numeric2) repalloc(n, calculeteNumericSize(newNumberOfComponents));
	}else {
		oldSize = getNumericValuesLength(n);
		out = (numeric2) repalloc(n, calculeteNumericSize(newNumberOfComponents));
		//memset(&(getNumericValues(n)[oldSize]), 0, newNumberOfComponents*sizeof(numeric2cmp) - oldSize);
	}

	out->size = calculeteNumericSize(newNumberOfComponents);
	return out;
}

/** Converts numeric2 type to char array with specified radix
 * @param n numeric2 value to convert.
 * @param radix radix to use.
 * @return null teminated numeric value in specified radix.
 */
Datum numeric2_out(PG_FUNCTION_ARGS) 
{
	fprintf(stderr, "In function out");
	PG_RETURN_CSTRING("aaa");

	numeric2 numeric = (numeric2) PG_GETARG_POINTER(0);
	int radix = 10;
	numeric2 n = numeric2_clone(numeric);
	int result_tmp_size;
	char* result_tmp;
	int pos;
	int i;
	int writen = 0;
	numeric2cmp* values;
	int valuesLength;

	PG_RETURN_CSTRING("aaaa");

	if (radix < 0 || radix > 16) {
		//error
		return (char*) 0;
	};

	result_tmp_size = getNumericValuesLength(n) * 10 * sizeof(char) + 2 + getNumeric2Sign(n);
	result_tmp = (char*) malloc(result_tmp_size); //Too big
	result_tmp[result_tmp_size - 1] = '\0';
	pos = result_tmp_size - 2;

	values = getNumericValues(n);
	valuesLength = getNumericValuesLength(n);

	while (!numeric2_is_zero(n)) {
		numeric2cmp reminder = __numeric2_div_and_remineder_by_component(getNumericValues(n), getNumericValuesLength(n), 10);			
		result_tmp[pos--] = radix_characters[reminder];
		writen++;

		if (writen == getNumeric2Scale(n))
			result_tmp[pos--] = '.';
	}

	if (writen == 0) {
		result_tmp[pos--] = '0';
	}else {
		if (result_tmp[pos+1] == '.')
			result_tmp[pos--] = '0';

		if (getNumeric2Sign(n))
			result_tmp[pos--] = '-';
	}
	pfree(n);
	PG_RETURN_CSTRING("aaaa");
}

Datum _numeric2_in(char *str);

Datum numeric2_in(PG_FUNCTION_ARGS) {
	numeric2 out = _numeric2_in(PG_GETARG_CSTRING(0));
	fprintf(stderr, "Created number with number of components %d, size %d\n", getNumericValuesLength(out), out->size);
	fprintf(stderr, "When casting then %d\n", out[0]);
	PG_RETURN_NUMERIC2(out);
}

Datum _numeric2_in(char *str) {
	fprintf(stderr, "In function in\n");
	//char* str = PG_GETARG_CSTRING(0);
	int i = 0;
	short sign;
	short scale = 0;
	int size = strlen(str);
	numeric2 out = numeric2_allocate(1, 0, 0);
	PG_RETURN_NUMERIC2(out);
	
	if (size == 0) {
		PG_RETURN_POINTER(out);
	}

	if (str[i] == '-') {
		sign = 0x8000;
		i++;
	}else {
		sign = 0;
	}

	for (; i < size; i++) {
		if (str[i] == '.') {
			scale = size - i - 1;
		}else {
			out = __numeric2_multiply_by_component(out, 10);
			out = __numeric2_add_component(out, str[i] - '0');
		}
	}

	out->scale = sign | (scale & 0x7fff);
	//fprintf(stderr, "Created number with number of components %d and size %d \n", getNumericValuesLength(out), out->size);
	PG_RETURN_NUMERIC2(out);
}

numeric2 numeric2_add(numeric2 a, numeric2 b)
{
	int outSize;
	numeric2 result, out;
	short scaleA, scaleB;
	unsigned int carry = 0;
	int i;
	unsigned long long int  x, y;
	unsigned long long int v;

	//TODO: We don't check sign here is no substraction!!!
	if (getNumeric2Scale(a) != getNumeric2Scale(b)) {
		scaleA = getNumeric2Scale(a);
		scaleB = getNumeric2Scale(b);

		if (scaleA < scaleB) {
			out = numeric2_clone(a);
			for (i=scaleA; i < scaleB; i++) 
				out = __numeric2_multiply_by_component(out, 10);
			out->scale = getNumeric2Sign(out) | (scaleB & 0x7fff);

			result = numeric2_add(out, b);
			pfree(out);
			return result;
		}else {
			out = numeric2_clone(b);
			for (i=scaleB; i < scaleA; i++)
				out = __numeric2_multiply_by_component(out, 10);

			out->scale = getNumeric2Sign(out) | (scaleA & 0x7fff);
			result = numeric2_add(a, out);
			pfree(out);
			return result;
		}
	}

	outSize = (getNumericValuesLength(a) > getNumericValuesLength(b) ? getNumericValuesLength(a) : getNumericValuesLength(b));
	result = numeric2_allocate(outSize + 1, 1, a->scale);

	//TODO Split this into two for if lenghts are different
	for (i = 0; i < outSize; i++) {
		x = (i < getNumericValuesLength(a)) ? getNumericComponentAt(a, i) : 0;
		y = (i < getNumericValuesLength(b)) ? getNumericComponentAt(b, i) : 0;

		v = x + y + carry;
		carry = (v >> 32);
		getNumericValues(result)[i] = (unsigned int) v;
	}

	getNumericValues(result)[outSize] = carry;
	numeric2_trim_leading(result);

	return result;
}

numeric2 __numeric2_add_component(numeric2 a, numeric2cmp comp) {
	unsigned long long int _comp = comp;
	unsigned long long int v;
	unsigned int carry = 0;
	short size = getNumericValuesLength(a);
	numeric2cmp* values = getNumericValues(a);
	int i;

	for (i=0; i < size; i++) {
		v = _comp + values[i];
		values[i] = (numeric2cmp) (v);
		carry = (v >> 32);
	}

	if (carry != 0) {
		a = numeric2_reallocate(a, size + 1);
		getNumericValues(a)[size] = carry;
	}

	return a;
}

numeric2 numeric2_div_component(numeric2 n, numeric2cmp component, short targetScale) {
	numeric2 out;
	if (targetScale > getNumeric2Scale(n)) {
		out = __numeric2_multiply_by_component(n, (targetScale - getNumeric2Scale(n))*10);		
		out->scale = out->scale | (targetScale & 0x7fff);
	}else {
		out = numeric2_clone(n);
	}

	__numeric2_div_and_remineder_by_component(getNumericValues(out), getNumericValuesLength(out), component);
	numeric2_trim_leading(out);

	return out;
}

numeric2 __numeric2_multiply_by_component(numeric2 n, numeric2cmp component) {
	int i;
	
	int lastNonZeroValue = 0;
	short sourceSize = getNumericValuesLength(n);

	numeric2cmp* values = getNumericValues(n);

	unsigned long long int v;
	unsigned long long int carry = 0;

	for (i=0; i < sourceSize; i++) {
		v = values[i];
		v = v * component + carry;
		values[i] = v & 0xffffffff;
		if (values[i] != 0)
			lastNonZeroValue = i;
		carry = (v >> 32);
	}

	if (carry != 0) {
		n = numeric2_reallocate(n, sourceSize + 1);
		getNumericValues(n)[sourceSize] = carry;
	}

	return n;
}

numeric2 __numeric2_negate(numeric2 n) {
	short sign = getNumeric2Sign(n);
	sign = ~sign;
	sign = sign & ((short) 0x8000);
	n->scale = getNumeric2Scale(n) | sign;
}

numeric2cmp __numeric2_div_and_remineder_by_component(numeric2cmp* values, numeric2cmp size, numeric2cmp divider) {
	int i;
	numeric2cmp reminder = 0;
	unsigned long long int v;

	for (i=size-1; i > -1; i--) {
		v = reminder;
		v = (v << 32) | values[i];
		reminder = v % divider;
		values[i] = v / divider;
	}

	return reminder;
}

numeric2 numeric2_clone(numeric2 src) {
	numeric2 result = numeric2_allocate(getNumericValuesLength(src), src->precision, src->scale);
	memcpy(getNumericValues(result), getNumericValues(src), getNumericValuesLength(src)*sizeof(numeric2cmp));

	return result;
}

int numeric2_is_zero(numeric2 n) {
	int i;
	for (i=getNumericValuesLength(n) - 1; i > -1; i--)
		if (getNumericValues(n)[i] != 0)
			return 0;
	return 1;
}

numeric2 numeric2_trim_leading(numeric2 n) {
	int length = getNumericValuesLength(n);
	int lastNonZeroValue = length - 1;

	for (; lastNonZeroValue > -1; lastNonZeroValue--)
		if (getNumericValues(n)[lastNonZeroValue] != 0)
			break;

	if (lastNonZeroValue != (length-1)) { //Last non zero value is different the last element 
		n = (numeric2) realloc(n, (lastNonZeroValue + 1) * sizeof(numeric2cmp) + sizeof(_numeric2));
		n->size = (lastNonZeroValue + 1) * sizeof(numeric2cmp) + sizeof(_numeric2);
	}

	return n;
}

