#ifndef __NUMERIC2_H__
#define __NUMERIC2_H__

#include <postgres.h>

#define DatumGetNumeric2(X)                ((numeric2) PG_DETOAST_DATUM(X))
#define DatumGetNumericCopy(X)    ((numeric2) PG_DETOAST_DATUM_COPY(X))
#define Numeric2GetDatum(X)                PointerGetDatum(X)
#define PG_GETARG_NUMERIC2(n)      DatumGetNumeric2(PG_GETARG_DATUM(n))
#define PG_GETARG_NUMERIC_COPY2(n) DatumGetNumeric2Copy(PG_GETARG_DATUM(n))
#define PG_RETURN_NUMERIC2(x)      return Numeric2GetDatum(x)

typedef unsigned int numeric2cmp;

typedef struct t_numeric2 {
	int32 size;
	short precision; //Should we use precision or should we calculate it?
	short scale;
} *numeric2, _numeric2;

#define getNumericValues(n) ((numeric2cmp*) ((void*) n + sizeof(_numeric2)))
#define getNumericValuesLength(n) ((n->size - sizeof(_numeric2))/sizeof(numeric2cmp))
#define getNumericComponentAt(n, i) (getNumericValues(n)[i])
#define NUMERIC2_MAX_COMPONENT = 0xffffffff;

#define getNumeric2Sign(n) ((n->scale & 0x8000) >> 15) //0 - positive, 1 negative
#define getNumeric2Scale(n) (n->scale & 0x7fff)

numeric2 numeric2_allocate(short numberOfComponents, short precision, short scale);
numeric2 numeric2_reallocate(numeric2 n, short newNumberOfComponents);

PG_FUNCTION_INFO_V1(numeric2_in);
PG_FUNCTION_INFO_V1(numeric2_out);
/*extern Datum
numeric2_out(PG_FUNCTION_ARGS);

extern Datum
numeric2_in(PG_FUNCTION_ARGS);*/

numeric2 numeric2_add(numeric2 a, numeric2 b);

/** Adds specified component to numeric. This method modifies original numeric {@code a} */
numeric2 __numeric2_add_component(numeric2 a, numeric2cmp comp);

numeric2 numeric2_div_component(numeric2 n, numeric2cmp component, short targetScale);

numeric2 __numeric2_negate(numeric2 n);

numeric2 numeric2_clone(numeric2 src);

/** Multiplies specified numeric by component. This method modifies original numeric {@code a}.
 * @return result (can be reallocated if size of numeric changed).
*/
numeric2 __numeric2_multiply_by_component(numeric2 n, numeric2cmp component);
/** Internal functions for dividing numeric with remineder by specified value 
 *
 * @param values - this is input and output
 * @return remined
*/
numeric2cmp __numeric2_div_and_remineder_by_component(numeric2cmp* values, numeric2cmp size, numeric2cmp divider);

numeric2 numeric2_trim_leading(numeric2 n);

int numeric2_is_zero(numeric2 n);
#endif
